#include "myIncludes.h"
#include "myProtos.h"
#include "Sound.h"
#include "stdio.h"

extern long			gmyTix[20];
extern DialogPtr	gDlog;

OSErr get_music(void)
{
Point	pt;
SFReply	reply;
SFTypeList l;
OSErr	err;
short	fref, i;
Size	sz;
Ptr		datap, traverse;
Handle	h;
Rect	r;
char	ch;

	pt.h = pt.v = 50;
	l[0] = 'TEXT';
	l[1] = l[2] = l[3] = 0L;
	SFGetFile(pt, "\p", NULL, 1, &l, NULL, &reply);
	if ( ! reply.good )
		return(abortErr);

	GetDItem(gDlog, 2, &i, &h, &r);
	SetIText(h, reply.fName);
	
/* read file */
	err = FSOpen( reply.fName, reply.vRefNum, &fref );
	if (err)
		return(err);

	if ( err = SetFPos(fref, fsFromLEOF, 0L) )
		{
		FSClose(fref);
		return(err);
		}
	if ( err = GetFPos(fref, &sz) )
		{
		FSClose(fref);
		return(err);
		}
	if ( err = SetFPos(fref, fsFromStart, 0L) )
		{
		FSClose(fref);
		return(err);
		}
	datap = NewPtr(sz+1);
	datap[sz] = 0;	/* terminate */
	if ( (err = MemError()) || datap == NULL )
		return(err?err:-1);
	
	if ( err = FSRead(fref, &sz, datap) )
		{
		DisposPtr(datap);
		FSClose(fref);
		return(err);
		}

	FSClose(fref);
	
/* assign to global */
	
	for (i = 0; i < 1024; i++)
		{
		gTrebMusic[i].midiFreq = 0;
		gTrebMusic[i].duration = 0;
		gBassMusic[i].midiFreq = 0;
		gBassMusic[i].duration = 0;
		}
	
	gNumTrebNotes = 0;
	for (traverse = datap; (traverse < datap + sz) && (*traverse != ';'); traverse++)
		{
		sscanf(traverse, "%d %d", &gTrebMusic[gNumTrebNotes].midiFreq, &gTrebMusic[gNumTrebNotes].duration);
		gNumTrebNotes++;
		while ( (*traverse != ',') && (*traverse != ';') )
			traverse++;
		
		if (*traverse == ';')
			traverse--;	/* account for 'traverse++' above */
		}
	
	gNumBassNotes = 0;
	if ( (*traverse == ';') && (traverse < datap + sz) )
		{
		for (traverse++; traverse < datap + sz; traverse++)
			{
			sscanf(traverse, "%d %d", &gBassMusic[gNumBassNotes].midiFreq, &gBassMusic[gNumBassNotes].duration);
			gNumBassNotes++;
			while (*traverse != ',')
				traverse++;
			}
		}
	
	DisposePtr(datap);

/* determine members needed */
	for ( i = 0; i < 256; ++i )
		gMidiPart[i] = -1;
		
	gNumMembers = 0;
	for (i = 0; i < gNumTrebNotes; ++i)
		if (gMidiPart[gTrebMusic[i].midiFreq] < 0)
			gMidiPart[gTrebMusic[i].midiFreq] = gNumMembers++;
		
	for (i = 0; i < gNumBassNotes; ++i)
		if (gMidiPart[gBassMusic[i].midiFreq] < 0)
			gMidiPart[gBassMusic[i].midiFreq] = gNumMembers++;
	
	return(noErr);
}


OSErr send_music()
{
long		curMember, curNote, ticks, startTime;
AppleEvent	theEvent;
Ptr			trebp, bassp, curtrebp, curbassp;
OSErr		err;
SndCommand	snd;

	startTime = TickCount() + (gNumMembers * 2 * 60);	/* 2 seconds for each member */
	
	for (curMember = 0; curMember < gNumMembers; curMember++)
		{
		err = AECreateAppleEvent(kChoirEventClass, STEP2, gMembers[curMember],
			kAutoGenerateReturnID, kAnyTransactionID, &theEvent);
		if (err)
			return(err);
		
		trebp = NewPtrClear( gNumTrebNotes * sizeof(SndCommand) );
		bassp = NewPtrClear( gNumBassNotes * sizeof(SndCommand) );
		if ( (err = MemError()) || (trebp == NULL) || (bassp == NULL) )
			return(err?err:-1);
		
		curtrebp = trebp;
		for (curNote = 0; curNote < gNumTrebNotes; curNote++)
			{
			snd.param1 = gTrebMusic[curNote].duration;
			if ((gMidiPart[gTrebMusic[curNote].midiFreq] % gNumMembers) == curMember)
				{
				snd.param2 = gTrebMusic[curNote].midiFreq;
				if (snd.param2 == 0)
					snd.cmd = restCmd;
				else
					snd.cmd = freqDurationCmd;
				}
			else
				{
				snd.param2 = 0L;
				snd.cmd = restCmd;
				}
			
			copy_mem(&snd, curtrebp, sizeof(SndCommand));
			curtrebp += sizeof(SndCommand);
			}
			
		curbassp = bassp;
		for (curNote = 0; curNote < gNumBassNotes; curNote++)
			{
			snd.param1 = gBassMusic[curNote].duration;
			if ((gMidiPart[gBassMusic[curNote].midiFreq] % gNumMembers) == curMember)
				{
				snd.param2 = gBassMusic[curNote].midiFreq;
				if (snd.param2 == 0)
					snd.cmd = restCmd;
				else
					snd.cmd = freqDurationCmd;
				}
			else
				{
				snd.param2 = 0L;
				snd.cmd = restCmd;
				}
			
			copy_mem(&snd, curbassp, sizeof(SndCommand));
			curbassp += sizeof(SndCommand);
			}
		
		ticks = startTime + gTix[curMember] - gmyTix[curMember];
		
		err = AEPutParamPtr(&theEvent, TICKS_PARAM, typeLongInteger, (Ptr)&ticks, 4);
		if (err == noErr)
			err = AEPutParamPtr(&theEvent, MUSIC_TREBLE, TREBLE_VOICE, trebp, gNumTrebNotes*sizeof(SndCommand));
		if (err == noErr)
			err = AEPutParamPtr(&theEvent, MUSIC_BASS, BASS_VOICE, bassp, gNumBassNotes*sizeof(SndCommand));
		if (err == noErr)
			err = AESend(&theEvent, NULL, kAENoReply + kAECanInteract + kAECanSwitchLayer,
					kAEHighPriority, kAEDefaultTimeout*10, NULL, NULL);
		
		if (err)
			SysBeep(0);
		err = AEDisposeDesc(&theEvent);
		}
	
	DisposePtr(trebp);
	DisposePtr(bassp);
	return(noErr);
}
